;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; PROJECT: Carnegie Mellon Planetary Rover Project
;;;          Task Control Architecture 
;;; 
;;; MODULE: tca - lisp
;;;
;;; FILE: .lisp
;;;
;;; ABSTRACT:
;;; This file is tcaTrace.lisp
;;; It contains code to parse a TCA log and recreate a version of the task
;;; tree.  This code was modified from tcaTimings.lisp.
;;; The log should be created (at a minimum) with the options -Lmt
;;;  (although adding the "i" option, which ignores logging registration
;;;    messages, is recommended).
;;;
;;; $Revision: 1.3 $
;;; $Date: 1995/01/24 23:59:26 $
;;; $Author: rich $
;;;
;;; REVISION HISTORY
;;;
;;;  $Log: tcaTrace.lisp,v $
;;; Revision 1.3  1995/01/24  23:59:26  rich
;;; Release of tca 7.9.  Mostly speed improvements.
;;; The cvs binaries may now be located in /usr/local.
;;; Reformatting to less than 80 character lines.
;;;
;;; Revision 1.2  1993/06/13  23:26:20  rich
;;; Added comments at the front of each file.
;;; Fixed prefix for log entries.
;;; Made changes needed for vx works style global variables.
;;;
;;;  4-Jun-90 Reid Simmons at CMU
;;;           Created.
;;;
;;;  25-Nov-92 Richard Goodwin at CMU
;;;           Modified to keep track of sub messages and block types.
;;;           This code is now used by other modules that manipulate the
;;;           information stored in the *modules* data structure.
;;;           timingreport.lisp, and tcaGoalTree.lisp.  The display routines
;;;           have been left in, but they won't work properly.
;;;

#+IBM-RT-PC (setq extensions:*load-if-source-newer* :COMPILE)

;;; The module structure keeps track of the state of a single module in 
;;; the tca system.  There is one per active module when analyzing a trace.
;;; The only module that is slightly different is the *central-module*.  It
;;; has the same structure, but different sematics.  The *central-module* does
;;; not "process" commands, goals or querries.  It can originate querries and
;;; is the place where 
(defstruct (module)
  (name "" :type string)   ;; module name 
  (inactive NIL :type list)  ;; list of messages that are inactive
  (pending NIL :type list)  ;; list of messages that are pending
  (active NIL :type list)  ;; list of active messages
  (complete NIL :type list)  ;; list of messages that have completed.
  (killed NIL :type list)  ;; list of killed messages
  (bypass nil :type list) ;; list of exception messages being bypassed.
  )

;;; A block corresponds to a message/response pair.  
(defstruct (block (:print-function block-printer))
  
  ;; identification slots
  (name "" :type string)
  (refId 0 :type integer) ;; tca reference id. Unique only while message alive.
  
  ;; Time slots.
  (created 0 :type integer)  ;; when message was first created
  (pending 0 :type integer) ;; when message was ready to be handled
  (start 0 :type integer) ;; when message was issued (start of handling)
  (end NIL :type (or NULL integer)) ;; when message was handled.
  (killed 0 :type integer) ;; when message was handled.
  
  ;; tree slots.
  (subBlocks nil :type list)  ;; list of children
  (superBlock NIL :type block) ;; pointer to parent block.
  
  ;; status slots
  ;; status is one of ACTIVE RESUMPTION PENDING SUCCESS 
  ;; INACTIVE TOP-LEVEL-QUERY
  (status NIL :type symbol) 
  (kind NIL :type symbol) ;; one of : COMMAND, GOAL, EXCEPTION, QUERY
  (source nil :type module) ;; module that sent the message.
  (dest nil :type module)  ;; module that received the message.
  (topLevel NIL :type symbol)  ;; Is the message a top level message?
  (waitingFor "" :type string) ;; name of query message is waiting for, if any
  (waitingForId 0 :type integer) ;; id of query message is waiting for, if any
  
  ;; data slots 
  (data nil )
  )

(defun block-printer (block stream arg3)
  (declare (ignore arg3))
  (format stream 
	  "#<~A~A ~A{~A}: ~A --> ~A ~A Start: ~D End: ~D Sub-Blocks : ~D>#"
	  (if (block-topLevel block)
	      "Top Level "
	    "")
	  (block-kind block)
	  (block-name block)
	  (block-refId block)
	  (if (block-source block)
	      (module-name (block-source block)))
	  (if (block-dest block)
	      (module-name (block-dest block)))
	  (block-status block)
	  (block-start block)
	  (block-end block)
	  (length (block-subBlocks block))))

;;; timings : structure used to record information about a timing histogram.
(defstruct (timings (:print-function timings-printer))
  file
  start
  (duration 0)
  modules)

(defun timings-printer (timings stream arg3)
  (declare (ignore arg3))
  (format stream "#<TIMINGS ~A: ~D Modules, ~D Messages, ~A Duration>#"
	  (timings-file timings)
	  (length (timings-modules timings)) 
	  (count-messages (timings-modules timings))
	  (time-string (timings-duration timings))))

;;; Define a types table for mapping the key word at the beginning of a line
;;; to a symbol representing the type of the line.  Lines that don't begin
;;; with one of the keywords are ignored.
(defparameter *types* '(("Query" . QUERY) 
			("Goal" . GOAL) 
			("Command" . COMMAND)
			("Constraint" . CONTRAINT)
			("Exception" . EXCEPTION)
			("Killed" . KILLED)
			("IMonitor" . INTERVALMONITOR)
			("PMonitor" . POLLINGMONITOR)
			("FireDemon" . FIREDEMON)
			("  Reply" . REPLY) 
			("  Failure" . FAILURE)
			("  Success" . SUCCESS)
			("  ByPass" . BYPASS)
			("   Asserting" . ASSERTION)
			("Will" . WILLKILL)
			("Adding" . ADDCONTRAINT)
			("WARNING:" . WARNING)
			))

;;; *debug* flag to indicate if debugging information is to be printed.
(defparameter *debug* T)  

;;; save the default read table, and make one for parsing the data lines.
(defvar *default-read-table* (copy-readtable nil))
(defvar *data-read-table* nil)


;;; *statuses* : table for mapping status strings to symbols.
(defparameter *statuses* '(("Sent" . SENT) ("Inactive" . INACTIVE)
			   ("Pending" . PENDING)))

;;; *line-pos* : global variable indicating the current parse position in the
;;; current input line.
(defvar *line-pos* 0)

;;; *start-time* : set to the first time found in the trace log.
(defparameter *start-time* 0)

;;; *end-time* : set to the last time found in the trace log.
(defparameter *end-time* 0)

;;; *central-module-name* : the name of the central module.
(defconstant *central-module-name* "TCA Server")

;;; *central-module* : a module structure for the central module, one of the 
;;; module structures in the *modules* list.
(defvar *central-module* nil)

;;; *on-hold* : string constant for messages not ready to be issued.
(defconstant *on-hold* "ON HOLD")

;;; Postscript output constants used for the histogram.
(defconstant *ppi* 72)

(defconstant *tick-length* 4)

(defconstant *pending-dec* 2)

(defconstant *wait-shade* .8)
(defconstant *busy-shade* .1)

(defconstant *page-height* 11.0)
(defconstant *page-width*   8.5)
(defconstant *page-border*  0.25)
(defconstant *overlap*      0.25)
(defconstant *usable-height* (- *page-height* *page-border* *page-border*))
(defconstant *usable-width* (- *page-width* *page-border* *page-border*))

(defconstant *left-margin* 5)


;;; *resource-messages* : A list of lists, each of the form 
;;; "((module resource-name) . message-names)"
;;;  where each of the messages is associated with the given resource.
(defparameter *resource-messages* NIL)

(defun resource-module (moduleName msgName)
  (dolist (resource-message *resource-messages*)
    (when (and (string-equal moduleName (first (first resource-message)))
	       (and (member msgName (cdr resource-message)
			    :TEST #'STRING-EQUAL)))
      (return (second (first resource-message))))))

(defun warning ( formatStr &rest params)
  (format T "~%WARNING: ")
  (apply #'format (cons T 
			(cons formatStr params)))
  )

(defun refEql (ref1 ref2)
  (or (null ref1)
      (null ref1)
      (eql ref1 ref2)))

(defun name-member (name list-of-names)
  (declare (inline name-member))
  (member name list-of-names :TEST #'STRING-EQUAL))

(defun find-module (moduleName moduleList)
  (car (member moduleName  moduleList :KEY #'MODULE-NAME 
	       :TEST #'STRING-EQUAL)))

(defun find-block (blockName blockList)
  (car (member blockName  blockList :KEY #'BLOCK-NAME :TEST #'STRING-EQUAL)))

(defun find-on-hold (msg refId moduleList)
  (dolist (mod moduleList)
    (dolist (b (module-inactive mod))
      (if (and (string-equal msg (block-name b))
	       (refEql refId (block-refId b)))
	  (return-from find-on-hold mod)))))

(defun get-module (moduleName msg refId moduleList)
  (let ((realModuleName (or (resource-module moduleName msg) moduleName)))
    (if (string-equal realModuleName "ON HOLD")
	(or (find-on-hold msg refId moduleList)
	    *central-module*)
      (or
       (find-module realModuleName moduleList)
       (let ((module (make-module :NAME realModuleName)))
	 (nconc moduleList (list module))
	 module)))))

(defun initialize-timings ()
  (setq *start-time* NIL *end-time* 0)
  (setq *data-read-table* (copy-readtable *default-read-table*))
  (set-syntax-from-char #\: #\space *data-read-table* *data-read-table*)
  (set-syntax-from-char #\{ #\space *data-read-table* *data-read-table*)
  (set-syntax-from-char #\} #\space *data-read-table* *data-read-table*)
  (set-syntax-from-char #\> #\space *data-read-table* *data-read-table*)
  (set-syntax-from-char #\< #\space *data-read-table* *data-read-table*)
  (set-syntax-from-char #\, #\space *data-read-table* *data-read-table*)
  )

(defun resetReadTable ()
  (setf *readtable* (copy-readtable nil)))

(defun readline (file)
  (setq *line-pos* 0)
  (read-line file NIL))

(defun unparse-type (type)
  (setq *line-pos* 0)
  type)

(defun parse-type (line)
  (dolist (type *types*)
    (let* ((type-name (car type))
	   (type-length (length type-name)))
      (when (and (>= (length line) type-length)
		 (dotimes (i type-length T)
		   (unless (char-equal (aref line (+ i *line-pos*))
				       (aref type-name i))
		     (return NIL))))
	(incf *line-pos* type-length)
        (return (cdr type))))))

(defun space? (char) 
  (declare (inline space?))
  (char-equal char #\space))

;;; Some basic string editing commands
(defun deleteSubstr (string substr &optional (all t))
  (do ((pos (and (>= (length string) (length substr))
		0 )
	    (position (elt substr 1) string :start pos :test #'char-equal)))
      ((null pos) string)
    (cond ((null pos)
	   (return-from deleteSubstr string))
	  ((string-equal (subseq string pos (min (+ pos (length substr))
						 (length string)))
			 substr)
	   (setf string 
		 (concatenate 'string
			      (subseq string 0 (1- pos))
			      (subseq string 
				      (+ pos (length substr))
				      (length string))))
	   
	   (setf pos 0)
	   (if (not all)
	       (return-from deleteSubstr string)
	     ))
	  (t
	   (incf pos)))
    ))


;;;
;;; Starting at *line-pos*, find the string that starts after the last 
;;;   "start-char" and ends before the last "end-char".
;;; Sets *line-pos* to be the position of the "end-char" found.
;;;
(defun delimit-string (line start-char end-char)
  (let ((string-start (or (position-if-not #'(lambda (char) 
					       (char-equal char start-char))
					   line :START *line-pos*)
			  *line-pos*)))
    (setq *line-pos* (position end-char line :TEST #'char-equal
			       :START string-start))
    (if (null *line-pos*)
	(setq *line-pos* (length line)))
    (subseq line string-start *line-pos*)))

(defun parse-message-name (line)
  (let ((msg-name (delimit-string line #\space #\:))
	(refId nil)
	(name-len nil))
    ;; determine if there was a {refid}
    (when (setf name-len (position #\{ msg-name :test #'char-equal))
      (setf refId (read-from-string msg-name nil nil 
				    :start (1+ name-len)
				    :end (1- (length msg-name))))
      (setf msg-name (subseq msg-name 0 (1- name-len)))
      (if (eql refId -1)
	  (setf refId nil))
      )
    ;; Skip over the colon
    (incf *line-pos*)
    (values msg-name refid)))

;;;
;;; Module name might be several words long.
;;; End of module name is indicated by a space followed by a non-alphabetic
;;;   character.
;;; This could cause problems if people start their module names with 
;;;   non-alphabetic characters.
;;;
;;; Returns the module name (a string) and T or NIL, depending on whether
;;;   the module name begins with the word "Resource" 
;;;
(defun parse-module-name (line)
  (if (<= (1- (length line)) *line-pos*)
      (return-from parse-module-name nil))
  ;; Skip over spaces
  (setq *line-pos* (position-if-not #'space? line :START *line-pos*))
  (if (<= (1- (length line)) *line-pos*)
      (return-from parse-module-name nil))
  (do ((module-name "")
       (spacer "" " "))
      ;; Is there another word to the module name?
      ((or (and (space? (aref line *line-pos*))
	    (not (alpha-char-p (aref line (1+ *line-pos*)))))
	   (<= (1- (length line)) *line-pos* ))
       (let ((start (mismatch "Resource" module-name :TEST #'char-equal)))
	 (cond ((= start 8) (values (subseq module-name (1+ start)) T))
	       (T module-name))))
    (setq module-name (concatenate 'STRING module-name spacer
				   (delimit-string line #\space #\space)))))

(defun parse-source (line)
  (multiple-value-bind (name source-type?) (parse-module-name line)
    ;; Skip over " --> "
    (incf *line-pos* 5)
    (values name source-type?)))

(defun parse-dest (line)
  (parse-module-name line))

(defun parse-status (line)
  (if (<= (1- (length line)) *line-pos*)
      (return-from parse-status nil))
  (setq *line-pos* (1+ (position #\( line
				 :START *line-pos* :TEST #'char-equal)))
  (let ((status-string (delimit-string line #\( #\))))
    ;; Skip over trailing parenthesis
    (incf *line-pos*)
    (dolist (status *statuses* (error "~%Unknown status ~A" status-string))
      (when (string-equal status-string (car status))
	(return (cdr status))))))

(defun parse-time (line)
  (if (<= (1- (length line)) *line-pos*)
      (return-from parse-time nil))
  (let ((hours 0)
	(minutes 0)
	(seconds 0)
	(msecs 0)
	(cursor 0))
    (multiple-value-setq (hours cursor) (parse-integer line :START *line-pos*
						       :JUNK-ALLOWED T))
    (multiple-value-setq (minutes cursor) (parse-integer line
							 :START (1+ cursor)
							 :JUNK-ALLOWED T))
    (multiple-value-setq (seconds cursor) (parse-integer line
							 :START (1+ cursor)
							 :JUNK-ALLOWED T))
    (multiple-value-setq (msecs cursor) (parse-integer line
						       :START (1+ cursor)
						       :JUNK-ALLOWED T))
    (setq *line-pos* cursor)
    (let ((time (+ msecs (* 100 (+ seconds (* 60 (+ minutes (* 60 hours))))))))
      (unless *start-time* (setq *start-time* time))
      (setq *end-time* time)
      time)))

(defun time-string (time-in-hundredths &optional ignore-hundredths)
  (multiple-value-bind (time-in-seconds hundredths) (floor time-in-hundredths
							   100)
    (multiple-value-bind (time-in-minutes seconds) (floor time-in-seconds 60)
      (multiple-value-bind (hours minutes) (floor time-in-minutes 60)
	(format NIL "~:[~D:~;~*~]~:[~2,'0D:~;~*~]~2,'0D~:[.~2,'0D~;~]"
		(zerop hours) hours (and (zerop minutes) (zerop hours)) minutes
		seconds ignore-hundredths hundredths)))))

;;; get the active block from the module.  If a mesage name is given, 
;;; search up the active stack for that message.
(defun active-block (module &optional (msg nil) (refId nil)) 
  (if (not (eql module "")) ;; use to be *central-module*
      (dolist (b (module-active module))
	(if (and (member (block-status b) 
			 '(ACTIVE WAIT-FOR-REPLY RESUMPTION TOP-LEVEL-QUERY))
		 (or (null msg)
		     (string-equal msg (block-name b)))
		 (refEql refId (block-refId b)))
	    (return-from active-block b)
	  )))
  nil)

;;; get the pending block from the module.  If a mesage name is given, 
;;; search up the pending stack for that message.
(defun pending-block (module &optional (msg nil) (refId nil)) 
  (dolist (b (module-pending module))
    (if (and (or (null msg)
		 (string-equal msg (block-name b)))
	     (refEql refId (block-refId b)))
	(return-from pending-block b)
      ))
  nil)

;; find the block being replied to given a module and a message name.
(defun findWaitingBlock (module msg refId) 
  (dolist (b (module-active module))
    (if (and (string-equal (block-waitingFor b) msg)
	     (refEql refId (block-waitingForId b))
	     (member (block-status b)
		     '(WAIT-FOR-REPLY TOP-LEVEL-QUERY)))
	(return-from findWaitingBlock b)))
  nil)

;;;
;;;
;;; If a message is sent, but no handler is around, TCA puts the message
;;;  on a resource by the name of the msg.  Once a handler is found, the
;;;  message is moved to the new resource and sent when things are all ready.
;;; This takes care of that situation.
;;;
(defun pending-awaiting-handler (msg refId resource-module)
  (let ((msg-resource (find-module msg *modules*)))
    (when msg-resource 
      (let ((pending-block (find-pending-block msg refId msg-resource)))
	(push pending-block (module-pending resource-module))
	(setf (module-pending msg-resource) 
	      (delete pending-block (module-pending msg-resource)))
	pending-block))))

(defun find-holding-block (msg refId module)
  (dolist (b (module-inactive module))
    (if (and (string-equal msg (block-name b))
	     (refEql refId (block-refId b)))
	(return-from find-holding-block b)))
  nil)

(defun find-pending-block (msg refId module)
  (dolist (b (module-pending module))
    (if (and (string-equal msg (block-name b))
	     (refEql refId (block-refId b)))
	(return-from find-pending-block b)))
  (pending-awaiting-handler msg refId module))

(defun find-complete-block (msg refId module)
  (dolist (b (module-complete module))
    (if (and (string-equal msg (block-name b))
	     (refEql refId (block-refId b)))
	(return-from find-complete-block b))))

(defun create-block (msg refId module stime status source 
			 from-pending? smodule kind)
  (let ((block nil))
    (cond ((and (eql smodule *central-module*)
		(string-equal source *on-hold*))
	   (when (null (setf block (find-holding-block msg refId smodule)))
	     (error "~% ON HOLD block not found ~A ~A" msg refId))
	   (setf (module-inactive smodule)
		 (delete block (module-inactive smodule)))
	   (if (null (block-dest block))
	       (setf (block-dest block) module))
	   )
	  (from-pending?
	   (if (null (setf block (find-pending-block msg refId module)))
	       (error "~%Pending block not found for ~A ~A" msg refId))
	   (setf (module-pending module)
		 (delete block (module-pending module)))
	   (if (null (block-dest block))
	       (setf (block-dest block) module))
	   )
	  (t (setf block (make-block :NAME msg :REFID refId
				     :CREATED stime :KIND kind
				     :source smodule 
				     :dest (if (not 
						(eql module *central-module*))
					       module)
				     ))
	     (cond ((block-source block)
		    (let ((active-block (active-block (block-source block))))
		      (cond (active-block
			     (push block (block-subBlocks active-block))
			     (push active-block (block-superBlock block))
			     )
			    (T (setf (block-topLevel block) T)
			       (when *debug*
				 (format t "~%Top Level Message ~A" 
					 block))))))
		   (T 
		    (error "~%Source Module not found for ~A ~A" msg refId))))
	  )
    (when (eql status 'INACTIVE)
      (setf (block-status block) 'INACTIVE)
      (push block (module-inactive module))
      )
    (when (eql status 'PENDING)
      (setf (block-status block) 'PENDING)
      (setf (block-pending block) stime)
      (push block (module-pending module))
      )
    (when (eql status 'SENT)
      (setf (block-status block) 'ACTIVE)
      (setf (block-start block) stime)
      (push block (module-active module))
      )
    (if (not (and (string-equal msg (block-name block))
		  (refEql refId (block-refId block))))
	(error "Names do not match ~A ~A" msg block))
    block))

(defun is-monitor? (moduleName)
  (declare (inline is-monitor?))
  (string-equal moduleName "*Monitor*"))

(defun process-query (line msg refId modules includeTopLevel)
  (multiple-value-bind (source is-source-resource?) (parse-source line)
    (let* ((dest (parse-dest line))
	   (status (parse-status line))
	   (qtime (parse-time line))
	   (smodule (unless (or is-source-resource? (is-monitor? source))
		      (get-module source msg refId modules)))
	   (active-block (if smodule (active-block smodule)))
	   (new-block (create-block msg refId 
				    (get-module dest msg refId modules) qtime
				    status source is-source-resource? 
				    smodule 'QUERY))
	   (dmodule (unless (is-monitor? dest) 
		      (get-module dest msg refId modules)))
	   )
      (when smodule
	(cond (active-block
	       (setf (block-end active-block) qtime)
	       (setf (block-status active-block) 'WAIT-FOR-REPLY)
	       (setf (block-waitingFor active-block) msg)
	       (setf (block-waitingForId active-block) refId)
	       (when *debug* (format t "~% Query ~A ~A by ~A ~A" 
				     msg
				     refId
				     (block-name active-block)
				     (block-refId active-block)
				     ))
	       )
	      (includeTopLevel
	       (setf active-block
		     (make-block :NAME msg :REFID refId
				 :STATUS 'TOP-LEVEL-QUERY
				 :CREATED qtime
				 :SUBBLOCKS (list new-block)
				 :START qtime :KIND 'QUERY
				 :SOURCE smodule
				 :DEST dmodule
				 :topLevel T
				 :waitingFor msg :waitingForId refId
				 ))
	       (push active-block (module-active smodule))
	       (setf (block-waitingFor active-block) msg)
	       (setf (block-waitingForId active-block) refId)
	       (push active-block (block-superblock new-block))
	       (setf (block-topLevel new-block) nil)
	       (when *debug*
		 (format t "~%Top Level Query ~A" active-block))
	       )))
      new-block
      )))

(defun process-exception (line msg refId modules type)
  (multiple-value-bind (source is-source-resource?) (parse-source line)
    (let* ((dest (parse-dest line))
	   (status (parse-status line))
	   (smodule (unless (or is-source-resource? (is-monitor? source))
		      (get-module source msg refId modules)))
	   (dmodule (unless (is-monitor? dest) 
		      (get-module dest msg refId modules)))
	   (active-block (if smodule (active-block smodule)))
	   (bypass (and smodule
			(car 
			 (member (list msg refId)
				 (module-bypass smodule)
				 :TEST #'(lambda (a b)
					   (and (string-equal (first a)
							      (block-name b))
						(or (null (second a))
						    (eql (second a)
							 (block-refId b)))))
				 ))))
	   )
      (cond (bypass
	     ;; just ignore the exception message, it is already being handled.
	     ;; this only works if the same module is handling both 
	     ;; the original and the bypass message.
	     (cond ((not (and (eql (block-source bypass)
				   smodule)
			      (eql (block-dest bypass)
				   dmodule)))
		    (warning "bypass not handled correctly ~A ~A" msg refId)
		    (create-block msg refId 
				  (get-module dest msg refId modules) 
				  (parse-time line)
				  status source is-source-resource?
				  smodule type)
		    )
		   (t
		    (when *debug*
		      (format t "~% Bypass retry ~A ~A" msg refId)) 
		    (setf (module-bypass smodule)
			  (delete bypass (module-bypass smodule)))
		    )))
	    (t 
	     (create-block msg refId
			   (get-module dest msg refId modules) 
			   (parse-time line)
			   status source is-source-resource? smodule type)
	     ))
      (when *debug*
	(format t "~% Exception ~A ~A by ~A ~A" msg refId
		(if active-block (block-name active-block)) status)
	)
      )))

(defun process-goal-command (line msg refId modules type)
  (multiple-value-bind (source is-source-resource?) (parse-source line)
    (let* ((dest (parse-dest line))
	   (status (parse-status line))
	   (smodule (unless (or is-source-resource? (is-monitor? source))
		      (get-module source msg refId modules)))
	   (active-block (if smodule (active-block smodule)))
	   (new-block 
	    (create-block msg refId
			  (get-module dest msg refId modules) (parse-time line)
			  status source is-source-resource? smodule type)))
      (when *debug*
	(format t "~% Goal ~A ~A by ~A ~A" msg refId
		(if active-block (block-name active-block)) status)
	)
      new-block
      )))

(defun parse-data (line)
  (let ((parse nil)
	(linePos 0)
	(lineLen (length line)))
    (if (or (string-equal line "    NO DATA")
	    (null line))
	(return-from parse-data nil))
    (setf *readtable* *data-read-table*)
    (setf line (deleteSubStr line "..."))
    (loop
     (multiple-value-bind (sym len) 
	 (read-from-string line nil nil :start linePos)
       (if sym (push sym parse)
	 (return-from parse-data (nreverse parse)))
       (setf linePos len)
       (when (>= linePos lineLen)
	 (setf *readtable* *default-read-table*)
	 (return-from parse-data (nreverse parse)))
       )
     )
    (setf *readtable* *default-read-table*)
    ))

(defun process-data (logfile block)
  (do ((line (readline logfile) (readline logfile)))
      ((unparse-type (parse-type line))
       (progn (setf (block-data block)
		    (nreverse (block-data block)))
	      line))
    (push (parse-data line) (block-data block))
    ;;(format t "~% DATA ~A ~% ~D: ~A" line 
    ;;    (length (first (block-data block)))
    ;;    (first (block-data block)))
    ))

(defun process-success (line msg refId modules)
  (unless (dolist (module modules)
	    (let ((active-block (active-block module msg refId)))
	      (when (and active-block 
			 (string-equal (block-name active-block) msg)
			 (or (null refId)
			     (eql (block-refId active-block) refId)))
		(setf (block-end active-block) (parse-time line))
		(setf (block-status active-block) 'SUCCESS)
		(let ((module (block-dest active-block)))
		  (when module
		    (setf (module-active module)  
			  (delete active-block (module-active module)))
		    (push active-block (module-complete module))))
		(when *debug* (format t "~% Success ~A ~A" msg refId))
		(return T))))
    (error "~%No corresponding goal/command message found for success ~A ~A" 
	   msg refId)))

(defun process-failure (line msg refId modules)
  (declare (ignore line))
  (unless (dolist (module modules)
	    (let ((active-block (active-block module msg refId)))
	      (when (and active-block 
			 (string-equal (block-name active-block) msg)
			 (refEql refId (block-refId active-block)))
		;; This is a hack -- should augment failure logging to print
		;;   time message is received.
		(setf (block-end active-block) *end-time*)
		(setf (block-status active-block) 'FAILURE)
		(let ((module (block-dest active-block)))
		  (when module
		    (setf (module-active module)  
			  (delete active-block (module-active module)))
		    (push active-block (module-complete module))))
		(when *debug*
		  (format t "~% Failure ~A ~A by ~A" msg refId
			  (block-name active-block)))
		(return T))))
    (error "~%No corresponding goal/command message found for failure ~A ~A" 
	   msg refId)))

(defun process-killed (line msg refId modules)
  (decf *line-pos* 2) ;; this is needed since there is no : after the msg name.
  (let ((inactive-block (find-on-hold msg refId modules))
	(qtime (parse-time line)))
    (when *debug*
      (format t "~% Bypass ~A ~A" msg refId)) 
    (dolist (module modules)
      (let ((pending-block (find-pending-block msg refId module))
	    (active-block (active-block module msg refId))
	    (complete-block (find-complete-block msg refId module))
	    )
	(cond 
	 (inactive-block
	  (push inactive-block (module-killed (block-dest inactive-block)))
	  (setf (block-killed inactive-block) qtime)
	  (setf (block-status inactive-block) 'KILLED)
	  (return-from process-killed inactive-block))
	 (pending-block
	  (push pending-block (module-killed module))
	  (setf (block-killed pending-block) qtime)
	  (setf (block-status pending-block) 'KILLED)
	  (setf (module-pending module) 
		(delete pending-block (module-pending module)))
	  (return-from process-killed pending-block))
	 (active-block 
	  (setf (block-killed active-block) qtime)
	  (setf (block-status active-block) 'KILLED)
	  (push active-block (module-killed module))
	  (setf (module-active module) 
		(delete active-block (module-active module)))
	  (return-from process-killed active-block))
	 (complete-block 
	  (setf (block-killed complete-block) qtime)
	  (setf (block-status complete-block) 'KILLED)
	  (push complete-block (module-killed module))
	  (setf (module-complete module) 
		(delete complete-block (module-complete module)))
	  (return-from process-killed complete-block))
	 )
	))
    nil))

(defun get-exception (msg refId modules)
  (dolist (m modules)
    (dolist (b (module-active m))
      (when (and (string-equal (block-name b) msg)
		 (refEql refId (block-refId b)))
	(return-from get-exception b))))
  nil)

(defun process-bypass (line msg refId modules)
  (declare (ignore line))
  ;; put the bypassed message on the bypass queue of the sending module
  ;; first find the exception being handled.  Hope there is only one.
  (let* ((exception (get-exception msg refId modules)))
    (when exception
      (push exception (module-bypass (block-source exception)))
      (when *debug*
	(format t "~% Bypass ~A ~A" msg refId)) 
      ))
  )

(defun process-reply (line msg refId modules)
  (let* ((source (parse-source line))
	 (dest (parse-dest line))
	 (etime (parse-time line))
	 (smodule (get-module source msg refId modules))
	 (dmodule (unless (is-monitor? dest) 
		    (get-module dest msg refId modules)))
	 (active-block (active-block smodule msg refId))
	 (waiting-block (when dmodule (findWaitingBlock dmodule msg refId))))
    (if (and (null active-block)
	     (not (eql smodule *central-module*)))
	(error "~%Module ~A not executing a ~A ~A query" source msg refId))
    (when active-block
      (setf (block-end active-block) etime)
      (setf (block-status active-block) 'SUCCESS)
      (when smodule
	(setf (module-active smodule)  
	      (delete active-block (module-active smodule)))
	(push active-block (module-complete smodule))
	))
    (cond (waiting-block
	   (setf (block-end waiting-block) etime)
	   (setf (block-waitingFor waiting-block) "")
	   (cond ((and (block-topLevel waiting-block)
		       (string-equal (block-name waiting-block)
				     msg)
		       (refEql refId (block-refId waiting-block)))
		  (setf (block-status waiting-block) 'SUCCESS)
		  (when dmodule
		    (setf (module-active dmodule)  
			  (delete waiting-block (module-active dmodule)))
		    (push waiting-block (module-complete dmodule)))
		  )
		 (T
		  (setf (block-status waiting-block) 'ACTIVE)))
	   )
	  (T (warning "No waiting block for ~A ~A in module ~A~%"
		      msg refId dest)))
    active-block))

(defparameter *modules* NIL)

(defun prepare-timings (file modules)
  (make-timings :FILE file :MODULES modules :START *start-time* 
		:DURATION (- *end-time* *start-time*)))

(defun count-messages (modules &optional messages-of-interest)
  (let ((num-messages 0))
    (dolist (module modules)
      (dolist (block (module-active module))
	(when (and (eql (block-status block) 'ACTIVE)
		   (or (null messages-of-interest)
		       (name-member (block-name block) messages-of-interest)))
	  (incf num-messages))))
    num-messages))

;;;
;;; Multiple resources is a list of lists, each sublist being of the form:
;;;   ((module-name resource-name) . message-names),
;;;  where each of the messages is associated with the given resource.
;;;
(defun process-line (line logfile ignoreMsgs includeTopLevel modules)
  (if (null line)
      (return-from process-line nil))
  (let ((type (parse-type line))
	(block nil))
    (when type
      (multiple-value-bind (msg-name refId) (parse-message-name line)
	(unless (name-member msg-name ignoreMsgs)
	  (case type
	    (QUERY
	     (setf block 
		   (process-query line msg-name refId modules includeTopLevel))
	     (process-line (process-data logfile block)
			   logfile ignoreMsgs includeTopLevel modules)
	     )
	    ((GOAL COMMAND)
	     (setf block 
		   (process-goal-command line msg-name refId modules type))
	     (process-line (process-data logfile block)
			   logfile ignoreMsgs includeTopLevel modules)
	     )
	    (EXCEPTION
	     (process-exception line msg-name refId modules type)
	     (process-line (readline logfile) 
			   logfile ignoreMsgs includeTopLevel modules)
	     )
	    (SUCCESS 
	     (process-success line msg-name refId modules)
	     (process-line (readline logfile) 
			   logfile ignoreMsgs includeTopLevel modules)
	     )
	    (FAILURE
	     (process-failure line msg-name refId modules)
	     (process-line (readline logfile) 
			   logfile ignoreMsgs includeTopLevel modules)
	     )
	    (KILLED
	     (process-killed line msg-name refId modules)
	     (process-line (readline logfile) 
			   logfile ignoreMsgs includeTopLevel modules)
	     )
	    (BYPASS
	     (process-bypass line msg-name refId modules)
	     (process-line (readline logfile) 
			   logfile ignoreMsgs includeTopLevel modules)
	     )
	    (REPLY
	     (setf block (process-reply line msg-name refId modules))
	     (if block
		 (process-line (process-data logfile block)
			       logfile ignoreMsgs includeTopLevel modules)
	       (process-line (readline logfile) 
			     logfile ignoreMsgs includeTopLevel modules))
	     )))))))

(defun analyze-timings (logfile-name
			&key ignoreMsgs resourceMsgs includeTopLevel)
  (let ((modules (list (make-module :NAME *central-module-name*))))
    (setq *modules* modules
	  *resource-messages* resourceMsgs
	  *central-module* (first modules))
    (initialize-timings)
    (with-open-file 
     (logfile logfile-name :DIRECTION :INPUT)
     (do ((line (readline logfile) (readline logfile)))
	 ((null line))
       (process-line line logfile ignoreMsgs includeTopLevel modules)
       )
     (prepare-timings logfile-name modules))))

;;;
;;; Options:
;;;   TITLE (string): Title of timings graph, displayed along top of
;;;      page.  Default is empty string.
;;;   ORIENTATION (:HORIZONTAL or :VERTICAL): Does time increase across the
;;;      horizontal or vertical axis.  Defaults to :HORIZONTAL.
;;;   HEIGHT (float): Height of the display, in inches.  Defaults to 10.5
;;;      for vertical orientation, 8.0 for horizontal orientation.
;;;   WIDTH (float): Width of the display, in inches.  Defaults to 8.0
;;;      for vertical orientation, 10.5 for horizontal orientation.
;;;   START (string): Do not display any messages that start before this time
;;;      (time is in same format as in TCA log -- hh:mm:ss.ss)
;;;   END (string): Do not display any messages that end after this time
;;;      (time is in same format as in TCA log -- hh:mm:ss.ss)
;;;   SECONDS-PER-TICK (integer): Number of seconds between each tick mark
;;;      along the time axis.  Defaults to 30 (every half minute).
;;;   SECONDS-PER-LABEL (integer): Number of seconds between each time label.
;;;      Defaults to 60 (every minute).
;;;   MODULES-TO-DISPLAY (list of strings): Modules whose timings should be 
;;;      displayed.  Modules are displayed in the order given.
;;;      Defaults to all modules except for "TCA Server" (ie, central).
;;;   LABEL-MODULES (integer, NIL or T): Method for displaying module names.
;;;        NIL:     Do not display
;;;        T:       Display complete name
;;;        integer: Display first N characters.
;;;      Defaults to T.
;;;   LABEL-MESSAGES (integer, NIL or T): Method for displaying message names.
;;;      Same semantics as for the "label-modules" keyword.
;;;      Defaults to NIL (ie, no message names displayed).
;;;
(defun display-timings (timings disp-file-name 
				&key (title "") (orientation :HORIZONTAL)
				(height (if (eql orientation :HORIZONTAL)
					    *usable-width* *usable-height*))
				(width (if (eql orientation :HORIZONTAL)
					   *usable-height* *usable-width*))
				start end (seconds-per-tick 30) (seconds-per-label 60)
				modules-to-display (label-modules T) label-messages)
  
  (let ((all-modules (timings-modules timings))
	modules)
    (setq modules     
	  (cond (modules-to-display
		 (mapcar #'(lambda (mod-name)
			     (let ((module (find-module mod-name all-modules)))
			       (unless module
				 (error "~%~A not a valid module name"
					mod-name))
			       module))
			 modules-to-display))
		(T (remove *central-module-name* all-modules
			   :KEY #'MODULE-NAME :TEST #'STRING-EQUAL))))
    
    (with-open-file (disp-file disp-file-name
			       :DIRECTION :OUTPUT :IF-EXISTS :SUPERSEDE)
		    (case orientation
		      (:HORIZONTAL (display-horizontal disp-file timings title
						       height width start
						       end seconds-per-tick
						       seconds-per-label
						       modules label-modules
						       label-messages))
		      (:VERTICAL (display-vertical disp-file timings title
						   height width start
						   end seconds-per-tick
						   seconds-per-label 
						   modules label-modules
						   label-messages))
		      (T (error "~%Unknown orientation: ~A" orientation)))
		    (format disp-file "~%showpage~%"))))

(defun add-postscript-header (file title height width)
  (format file "%!PS-Adobe-1.0~%%%Creator: Reid Simmons")
  (format file "~%%%TCA Timings Chart~%%%Title: ~A" title)
  (format file "~%%%BoundingBox ~D ~D ~D ~D" 
	  8 (floor (ppi .25)) (ceiling (ppi height)) (ceiling (ppi width)))
  (format file "~%/inch {~D mul} def" *ppi*)
  (format file "~%/TRfont {/Times-Roman findfont exch scalefont setfont} def")
  (format file "~%/YCenter % Stk: x, y, string~
          ~%   {3 1 roll moveto dup stringwidth -2 div 0 exch ~
          ~%    rmoveto pop show} def")
  (format file "~%/XCenter % Stk: x, y, string~
          ~%   {3 1 roll moveto dup stringwidth pop -2 div 0 ~
          ~%    rmoveto show} def")
  (format file "~%/RFlushYCenter % Stk: x, y, string~
          ~%   {3 1 roll moveto dup stringwidth -2 div exch neg exch~
          ~%    rmoveto show} def")
  (format file "~%/box     % Stk: xl, yb, dx, dy ~
          ~%   {newpath 4 2 roll moveto dup 0 exch rlineto exch 0 rlineto ~
          ~%    neg 0 exch rlineto closepath} def")
  (format file "~%/fillbox {setgray fill} def")
  (format file "~%/busy {box ~2F fillbox} def" *busy-shade*)
  (format file "~%/waiting {box ~2F fillbox} def" *wait-shade*)
  (format file "~%/enclose {box 0 setgray stroke} def"))

(defun ppi (inches) 
  (declare (inline ppi))
  (* inches *ppi*))

(defun display-title (file title left height width)
  (when (plusp (length title))
    (format file "~%24 TRfont~%~,1F ~,1F (~A) XCenter" 
	    (+ left (ppi (/ width 2))) (- (ppi height) 12) title)))

(defun display-axes (file top bottom left right horizontal?)
  (format file "~%% Display axes~%newpath ")
  (cond (horizontal?
	 (format file 
		 "~,1F ~,1F moveto ~,1F ~,1F lineto ~,1F ~,1F lineto stroke"
		 left top left bottom right bottom))
	(T (format file
		   "~,1F ~,1F moveto ~,1F ~,1F lineto ~,1F ~,1F lineto stroke"
		   right top left top left bottom))))

(defun display-temporal-axis (file x0 y0 num-points seconds-per-point
				   seconds-per-tick seconds-per-label horizontal?)
  (let ((points-per-tick (/ seconds-per-tick seconds-per-point))
	(points-per-label (/ seconds-per-label seconds-per-point)))
    
    (when (< points-per-tick 1) 
      (error "~%Need to have at least one point per tick"))
    
    (format file "~%% Display tick marks and time labels~
            ~%/tick % stk: x, y~%  {moveto ~:[~,1F 0~;0 ~,1F~] rlineto} def" 
	    horizontal? (- *tick-length*))
    
    (format file "~%newpath")
    (cond (horizontal? (do ((tick points-per-tick (+ tick points-per-tick)))
			   ((>= tick num-points))
		         (format file "~%~,1F ~,1F tick" (+ x0 tick) y0)))
	  (T (do ((tick num-points))
		 ((<= tick 0))
	       (decf tick points-per-tick)
	       (format file "~%~,1F ~,1F tick" x0 (+ y0 tick)))))
    (format file "~%stroke")
    
    (when (< points-per-label 1)
      (error "~%Need to have at least one point per label"))
    
    (format file "~%8 TRfont")
    (cond (horizontal?
	   (do ((label points-per-label (+ label points-per-label))
		(time seconds-per-label (+ time seconds-per-label)))
	       ((>= label num-points))
	     (format file "~%~,1F ~,1F (~A) XCenter"
		     (+ x0 label) (- y0 *tick-length* 8)
		     (time-string (* time 100) T))))
	  (T (do ((label num-points)
		  (time seconds-per-label (+ time seconds-per-label)))
		 ((<= label 0))
	       (decf label points-per-label)
	       (format file "~%~,1F ~,1F (~A) RFlushYCenter"
		       (- x0 *tick-length* 2) (+ y0 label)
		       (time-string (* time 100) T)))))))

;;;
;;; If limit is an integer, limit the maximum length of the name to that
;;;   number of characters.
;;;
(defun limited-name (name limit)
  (cond ((numberp limit)
	 (subseq name 0 (min limit (length name))))
	(T name)))

(defun display-module-name (file module x y label-modules horizontal?)
  (format file "~%~,1F ~,1F (~A) ~:[XCenter~;RFlushYCenter~]" 
	  x y (limited-name (module-name module) label-modules) horizontal?))

(defun display-modules (file modules-to-display module-axis
			     mod-axis-max mod-axis-min tstart tend seconds-per-point
			     label-modules label-messages horizontal?)
  (let* ((num-modules (length modules-to-display))
	 (mdelta (/ (- mod-axis-max mod-axis-min) (1+ num-modules)))
	 (mstart mod-axis-max))
    (dolist (module modules-to-display)
      (decf mstart mdelta)
      (when (or label-messages (eq module (first modules-to-display)))
	(format file "~%11 TRfont 0.2 setlinewidth"))
      (when label-modules
        (display-module-name 
	 file module (if horizontal? (- module-axis 2) mstart)
	 (if horizontal? mstart (+ module-axis 2))
	 label-modules horizontal?))
      (display-module-timings file module (if horizontal? module-axis mstart)
			      (if horizontal? (floor mstart) module-axis)
			      tstart tend (* seconds-per-point 100) (* mdelta .667)
			      label-messages horizontal?))))

(defun display-module-timings (file module x0 y0 timings-start timings-end
				    hundredths-per-point block-height
				    label-messages horizontal?)
  (let ((half-block (/ block-height 2))
	(last-active (car (last (module-active module))))
	mend)
    (when (module-pending module)
      (format file "~%% Draw pending messages~%newpath 2 setlinewidth")
      (dolist (pending (module-pending module))
	(when (block-end pending);; Might have been an aborted message
	  (when (and (>= (block-start pending) timings-start)
		     (<= (block-end pending) timings-end))
	    (let ((bstart (/ (- (block-start pending) timings-start)
			     hundredths-per-point))
		  (bend (/ (- (block-end pending) timings-start)
			   hundredths-per-point)))
	      (cond (horizontal?
		     (format file "~%~,2F ~,2F moveto ~,2F 0 rlineto" 
			     (+ X0 bstart) (- y0 half-block *pending-dec*)
			     (- bend bstart)))
		    (T (format file "~%~,2F ~,2F moveto 0 ~,2F rlineto"
			       (- X0 half-block *pending-dec*)
			       (- Y0 bstart) (- bstart bend))))))))
      (format file "~%stroke 0.2 setlinewidth"))
    
    (when label-messages (format file "~%6 TRfont"))
    (dolist (active (module-active module))
      (when (block-end active) ;; Might have been an aborted message
        (when (and (>= (block-start active) timings-start)
		   (<= (block-end active) timings-end))
	  (let ((bstart (/ (- (block-start active) timings-start)
			   hundredths-per-point))
		(bend (/ (- (block-end active) timings-start)
			 hundredths-per-point))
		(status (if (eql (block-status active) 'WAIT-FOR-REPLY)
			    "waiting" "busy")))
	    (cond (horizontal?
		   (format file "~%~,2F ~,1F ~,2F ~,1F ~A" 
			   (+ x0 bstart) (- y0 half-block) (- bend bstart)
			   block-height status))
		  (T (format file "~%~,1F ~,2F ~,1F ~,2F ~A"
			     (- x0 half-block) (- y0 bend) 
			     block-height (- bend bstart) status)))
	    (unless mend (setq mend bend))
	    (when (or (eql (block-status active) 'ACTIVE) (eq active last-active))
	      (cond (horizontal? (format file "~%~,2F ~,1F ~,2F ~,1F enclose"
					 (+ x0 bstart) (- y0 half-block)
					 (- mend bstart) block-height))
		    (T (format file "~%~,1F ~,2F ~,1F ~,2F enclose"
			       (- x0 half-block) (- y0 mend)
			       block-height (- mend bstart))))
	      (setq mend NIL)
	      (when label-messages
	        (cond (horizontal? 
		       (format file "~%~,2F ~,2F moveto (~A) show"
			       (+ x0 bstart) (+ y0 half-block *pending-dec*)
			       (limited-name (block-name active) label-messages)))
		      (T (format file "~%~,2F ~,2F moveto (~A) show"
				 (+ x0 half-block 2) (- y0 bstart 6)
				 (limited-name (block-name active) 
					       label-messages))))))))))))

(defun max-module-name (modules)
  (let ((max-name 0))
    (dolist (module modules)
      (setq max-name (max max-name (length (module-name module)))))
    max-name))

(defun display-temporal-extent (timings start end)
  (let ((display-start (cond (start (let ((*line-pos* 0))
				      (parse-time start)))
			     (T (timings-start timings))))
	(display-end (cond (end (let ((*line-pos* 0))
				  (parse-time end)))
			   (T (+ (timings-start timings)
				 (timings-duration timings))))))
    (values display-start display-end (- display-end display-start))))

(defun display-horizontal (file timings title height width start end 
				seconds-per-tick seconds-per-label modules-to-display
				label-modules label-messages)
  (let* ((modules (timings-modules timings))
	 (top (ppi (cond ((plusp (length title)) 
			  (- height 0.5))  ;; Half inch on top for title
			 (T height))))
	 (bottom (ppi 0.5));; Half inch for time labels
	 ;; Estimate 6 points per letter
	 (left (+ 2 (* 6 (cond ((null label-modules) 1)
			       ((numberp label-modules) label-modules)
			       (T (max-module-name modules))))))
	 (right (ppi width)))
    ;; Output R-L-T-B to make bounding box come out right for Scribe files
    (add-postscript-header file title height width)
    
    (format file "~%-90 rotate ~,1F inch 0 translate " (- *page-height*))
    
    ;; Start of double "for" loops to print multiple pages
    #|(format file "~%~,1F inch ~,1F inch ~,1F inch {"
	    (- *usable-width* height) (- *usable-width* *overlap*)
	    (- height *overlap*))
    (format file "~%~,1F inch ~,1F inch ~,1F inch {"
	    (- *page-border*) (- *usable-height* *overlap*)
	    (- width *overlap* *page-border*))|#
    
    (display-title file title left height width)
    (display-axes file top bottom left right T)
    (multiple-value-bind (display-start display-end display-duration)
	(display-temporal-extent timings start end)
      (let* ((num-points (- right left))
	     (seconds-per-point (/ display-duration 100 num-points)))
	(display-temporal-axis file left bottom num-points
			       seconds-per-point seconds-per-tick
			       seconds-per-label T)
	(display-modules file modules-to-display left top bottom
			 display-start display-end seconds-per-point
			 label-modules label-messages T)))
    
    ;; End of double "for" loops
    #|(format file "} for } for")|#
    ))

(defun display-vertical (file timings title height width start end
			      seconds-per-tick
			      seconds-per-label modules-to-display
			      label-modules label-messages)
  
  (let ((top (ppi (- height 0.5 0.25))) ;; Half inch on top for title plus
	;;   quarter inch for module names
	(bottom (ppi *page-border*))
	;; Estimate 25 points for time labels
	(left (+ 25 *tick-length* 2 (ppi *page-border*)))
	(right (ppi width)))
    
    ;; Start of double "for" loops to print multiple pages
    #|(format file "~%~,1F inch ~,1F inch ~,1F inch {"
	    (- *usable-width* height) (- *usable-width* *overlap*)
	    (- height *overlap*))
    (format file "~%~,1F inch ~,1F inch ~,1F inch {"
	    (- *page-border*) (- *usable-height* *overlap*)
	    (- width *overlap* *page-border*))|#
    
    (add-postscript-header file title height width)
    (display-title file title left height width)
    (display-axes file top bottom left right NIL)
    (multiple-value-bind (display-start display-end display-duration)
	(display-temporal-extent timings start end)
      (let* ((num-points (- top bottom))
	     (seconds-per-point (/ display-duration 100 num-points)))
	(display-temporal-axis file left bottom num-points seconds-per-point
			       seconds-per-tick seconds-per-label NIL)
	(display-modules file modules-to-display top right left
			 display-start display-end seconds-per-point label-modules
			 label-messages NIL)))
    
    ;; End of double "for" loops
    #|(format file "} for } for")|#
    ))

(defun print-centered-title (stream title topMsg)
  (when title
    (format stream "~%~VT~A~%" (/ (- (length topMsg) (length title)) 2) title))
  (format stream "~%~A" topMsg))

(defun report-utilization (timings &key file start end title)
  (cond (file
	 (with-open-file (output-file file :DIRECTION :OUTPUT :IF-EXISTS :SUPERSEDE)
			 (report-utilization1 output-file timings start end title)))
	(T (report-utilization1 T timings start end title))))

(defun report-utilization1 (stream timings start end title)
  (print-centered-title
   stream title
   "   MODULE     : ACTIVE NUM    SECS    %   PENDING NUM    SECS    %")
  (multiple-value-bind (display-start display-end duration)
      (display-temporal-extent timings start end)
    (let ((numMessages 0))
      (dolist (module (reverse (timings-modules timings)))
        (let ((active-time 0) (active-msgs 0)
	      (pending-time 0) (pending-msgs 0))
	  (dolist (block (module-active module))
	    (when (block-end block) ;; Might have been an aborted message
	      (when (and (>= (block-start block) display-start)
			 (<= (block-end block) display-end))
	        (when (eql (block-status block) 'ACTIVE) (incf active-msgs))
		(incf active-time (- (block-end block) (block-start block))))))
	  (dolist (block (module-pending module))
	    (when (block-end block)
	      (when (and (>= (block-start block) display-start)
			 (<= (block-end block) display-end))
		(incf pending-msgs)
		(incf pending-time (- (block-end block) (block-start block))))))
	  (incf numMessages active-msgs)
	  (unless (and (zerop active-time) (zerop pending-time))
	    (format stream
		    "~%~14A:        ~3D ~7,2F  ~4,1F          ~3D ~7,2F  ~4,1F"
		    (module-name module) active-msgs
		    (/ active-time 100) (* 100 (/ active-time duration))
		    pending-msgs (/ pending-time 100)
		    (* 100 (/ pending-time duration))))))
      (format stream "~2%     LOGGED ~D MESSAGES IN ~A~%" 
	      numMessages (time-string duration T)))))

(defstruct (message-report)
  name
  (number 0)
  (time 0.0)
  (waiting 0.0))

(defun report-messages (timings &key file start end title)
  (multiple-value-bind (display-start display-end duration)
      (display-temporal-extent timings start end)
    (let ((message-reports (create-message-reports timings
						   display-start display-end)))
      (cond (file
	     (with-open-file (output-file file 
					  :DIRECTION :OUTPUT :IF-EXISTS :SUPERSEDE)
			     (report-messages1 output-file message-reports duration title)))
	    (T (report-messages1 T message-reports duration title))))))

;;; The first on the list is a phony, so we can nconc with impunity
(defun get-message-report (msg message-reports)
  (or (car (member msg (cdr message-reports)
		   :TEST #'STRING-EQUAL :KEY #'message-report-name))
      (let ((message-report (make-message-report :NAME msg)))
	(nconc message-reports (list message-report))
	message-report)))

(defun create-message-reports (timings start end)
  (let ((message-reports (list NIL))
	current-message-report)
    (dolist (module (timings-modules timings))
      (setq current-message-report NIL)
      (dolist (block (reverse (module-active module)))
        (when (block-end block) ;; Might have been an aborted message
	  (when (and (>= (block-start block) start)
		     (<= (block-end block) end))
	    (case (block-status block)
	      (ACTIVE 
	       (setq current-message-report (get-message-report (block-name block)
								message-reports))
	       (incf (message-report-number current-message-report))
	       (incf (message-report-time current-message-report)
		     (- (block-end block) (block-start block))))
	      (WAIT-FOR-REPLY
	       (when current-message-report
		 (incf (message-report-waiting current-message-report)
		       (- (block-end block) (block-start block)))))
	      (RESUMPTION
	       (when current-message-report
		 (unless (string-equal (block-name block) 
				       (message-report-name current-message-report))
		   (warning "Current message is ~S, block is ~S~%"
			    (message-report-name current-message-report)
			    (block-name block)))
		 (incf (message-report-time current-message-report)
		       (- (block-end block) (block-start block))))))))))
    (cdr message-reports)))

(defun report-messages1 (stream message-reports duration title)
  (print-centered-title
   stream title
   "      MESSAGE       :  NUM TOTAL TIME   BUSY   (AVERAGE    %)  WAITING")
  (let ((numMessages 0))
    (dolist (message-report message-reports)
      (let ((number (message-report-number message-report))
	    (active-time (message-report-time message-report))
	    (waiting-time (message-report-waiting message-report)))
	(format stream "~%~20A: ~3D  ~7,2F   ~7,2F   (~5,2F   ~4,1F) ~7,2F"
		(message-report-name message-report) number 
		(/ (+ active-time waiting-time) 100)
		(/ active-time 100) (/ active-time 100 number)
		(* 100 (/ active-time duration)) (/ waiting-time 100))
	(incf numMessages number)))
    (format stream "~2%     LOGGED ~D MESSAGES IN ~A~%" 
	    numMessages (time-string duration T))))

(defun perform-checks ()
  (dolist (mod *modules*)
    (when (intersection (module-pending mod) (module-active mod))
      (format t "~% bad blocks ~A in ~% Module ~A" 
	      (intersection (module-pending mod) (module-active mod)) mod)
      (return-from perform-checks T))
    (when (intersection (module-pending mod) (module-complete mod))
      (format t "~% bad blocks ~A in ~% Module ~A" 
	      (intersection (module-pending mod) (module-complete mod)) mod)
      (return-from perform-checks T))
    (when (intersection (module-complete mod) (module-active mod))
      (format t "~% bad blocks ~A in ~% Module ~A" 
	      (intersection (module-complete mod) (module-active mod)) mod)
      (return-from perform-checks T))
    (dolist (a (module-pending mod))
      (when (not (eql (block-status a)
		      'PENDING))
	(format t "~% bad block ~A in ~% Module ~A" a mod)
	(return-from perform-checks a)))
    (dolist (a (module-active mod))
      (when (not (member (block-status a)
			 '(ACTIVE WAIT-FOR-REPLY TOP-LEVEL-QUERY)))
	(format t "~% bad block ~A in ~% Module ~A" a mod)
	(return-from perform-checks a)))
    (dolist (a (module-complete mod))
      (when (not (member (block-status a)
			 '(SUCCESS FAILURE)))
	(format t "~% bad block ~A in ~% Module ~A" a mod)
	(return-from perform-checks a)))
    )
  nil)

