/******************************************************************************
 *
 * PROJECT: Carnegie Mellon Planetary Rover Project
 *          Task Control Architecture 
 * 
 * (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
 * 
 * MODULE: communications
 *
 * FILE: com.c
 *
 * ABSTRACT:
 * 
 * Module Communications
 *
 * REVISION HISTORY 
 *
 * $Log: com.c,v $
 * Revision 1.8  1995/07/19  14:25:49  rich
 * Added display and dump to the central interface.
 * Fixed problem with direct querries not returning to the correct module.
 * Added Argv versions of provides and requires.
 *
 * Revision 1.7  1995/07/10  16:16:50  rich
 * Interm save.
 *
 * Revision 1.6  1995/07/08  17:50:54  rich
 * Linux Changes.  Also added GNUmakefile.defs.
 *
 * Revision 1.5  1995/07/06  21:15:44  rich
 * Solaris and Linux changes.
 *
 * Revision 1.4  1995/06/05  23:58:38  rich
 * Improve support of detecting broken pipes.  Add support for OSF 2.
 * Add return types to the global variable routines.
 *
 * Revision 1.3  1995/04/09  20:30:03  rich
 * Added /usr/local/include and /usr/local/lib to the paths for compiling
 * for sunOS machines. (Support for new vendor OS).
 * Create a tca directory in /tmp and put the socket in there so others can
 * delete dead sockets.  The /tmp directory has the sticky bit set so you
 * can't delete files even if you have write permission on the directory.
 * Fixes to libc.h to use the new declarations in the gcc header files and
 * avoid problems with dbmalloc.
 *
 * Revision 1.2  1995/04/07  05:02:50  rich
 * Fixed GNUmakefiles to find the release directory.
 * Cleaned up libc.h file for sgi and vxworks.  Moved all system includes
 * into libc.h
 * Got direct queries to work.
 * Fixed problem in allocating/initializing generic mats.
 * The direct flag (-c) now mostly works.  Connect message has been extended to
 * indicate when direct connections are the default.
 * Problem with failures on sunOS machines.
 * Fixed problem where tcaError would not print out its message if logging had
 * not been initialized.
 * Fixed another memory problem in modVar.c.
 * Fixed problems found in by sgi cc compiler.  Many type problems.
 *
 * Revision 1.1  1995/04/04  19:53:04  rich
 * Forgot com.h and com.c.
 *
 *
 * $Revision: 1.8 $
 * $Date: 1995/07/19 14:25:49 $
 * $Author: rich $
 *
 *****************************************************************************/

#include "globalM.h"

/******************************************************************************
 *
 * FUNCTION: BOOLEAN connectAtPort(const char *machine, int port, int *sd)
 *
 * DESCRIPTION: Connect at the TCP/IP connection.
 *
 * INPUTS: none.
 *
 * OUTPUTS: char *
 *
 *****************************************************************************/

BOOLEAN connectAtPort(const char *machine, int32 port, int *sd)
{
#ifndef VXWORKS
  struct hostent *hp;
#endif
  
  struct sockaddr_in server;
  
  bzero((char *)&server, sizeof(struct sockaddr_in));
  
  server.sin_family = AF_INET;
  server.sin_port = htons((u_short)port);
  
  if (isdigit(machine[0])) {
    /* printf("digit! \n");*/
    server.sin_addr.s_addr = inet_addr(machine);
  } else {
#ifndef VXWORKS
    if ((hp = gethostbyname(machine)) == NULL)
      return FALSE;
    BCOPY((char *)hp->h_addr, (char *)&server.sin_addr, hp->h_length);
#else
    if ((server.sin_addr.s_addr = hostGetByName(machine)) == ERROR)
      return FALSE;
#endif
  }
  
  if ((*sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return FALSE;
  
  /******* 17-aug-91: fedor: causes NeXT machine to fail connecting **
    if ((setsockopt(*sd, SOL_SOCKET, SO_REUSEADDR, (char *)0, 0)) < 0)
    return FALSE;
    *****/
  
  /******
    item = 0;
    if ((setsockopt(*sd, SOL_SOCKET, SO_LINGER, &item, sizeof(int))) < 0)
    return FALSE;
    *****/
  
#if !defined(applec) && !defined(__TURBOC__)
  { int item = 1;
    if ((setsockopt(*sd, SOL_SOCKET, SO_REUSEADDR, 
		    (char *)&item, sizeof(int))) < 0) {
      close(*sd);
      return FALSE;
    }
  }
#endif /* applec */
  
#ifndef VXWORKS
  { int value = TCP_SOCKET_BUFFER;
    if ((setsockopt(*sd, IPPROTO_TCP, TCP_NODELAY,
		    (char *)&value, sizeof(int))) < 0)
      return FALSE;
    
    setsockopt(*sd, SOL_SOCKET, SO_SNDBUF, (char *)&value, sizeof(int));
    setsockopt(*sd, SOL_SOCKET, SO_RCVBUF, (char *)&value, sizeof(int));
    
  }
#endif
  
  /* 24-Aug-90: fedor: VxWorks sockaddr_in is defined differently
     this should be checked to make a VxWorks version. */
  if (connect(*sd, (struct sockaddr *)&server, sizeof(server)) < 0) {
    close(*sd);
    return FALSE;
  }
  
  return TRUE;
}

/******************************************************************************
 *
 * FUNCTION: BOOLEAN connectAtSocket(const char *machine, int port, int *sd)
 *
 * DESCRIPTION: Connect at the unix socket.
 *
 * INPUTS: none.
 *
 * OUTPUTS: char *
 *
 *****************************************************************************/

BOOLEAN connectAtSocket(const char *machine, int32 port, int *sd)
{
  int value;
  struct sockaddr_un unix_server;
  
  bzero((char *)&unix_server, sizeof(struct sockaddr_un));
  unix_server.sun_family = AF_UNIX;
  sprintf(unix_server.sun_path,UNIX_SOCKET_NAME,port);
  
  if ((*sd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
    return FALSE;
  
  /******* 17-aug-91: fedor: causes NeXT machine to fail connecting **
    if ((setsockopt(*sd, SOL_SOCKET, SO_REUSEADDR, (char *)0, 0)) < 0)
    return FALSE;
    *****/
  
  /******
    item = 0;
    if ((setsockopt(*sd, SOL_SOCKET, SO_LINGER, &item, sizeof(int))) < 0)
    return FALSE;
    *****/
  
#ifndef VXWORKS
  value = UNIX_SOCKET_BUFFER;
  setsockopt(*sd, SOL_SOCKET, SO_SNDBUF, (char *)&value, sizeof(int));
  setsockopt(*sd, SOL_SOCKET, SO_RCVBUF, (char *)&value, sizeof(int));
#endif
  
  /* 24-Aug-90: fedor: VxWorks sockaddr_in is defined differently
   * this should be checked to make a VxWorks version.
   */
  if (connect(*sd, (struct sockaddr *)&unix_server, sizeof(unix_server)) < 0) {
    close(*sd);
    return FALSE;
  }
  
  return TRUE;
}
/******************************************************************************
 *
 * FUNCTION: Bolean connectAt(const char *machine, int port, int *sd)
 *
 * DESCRIPTION: Connect at the unix socket if possible, otherwise, at the 
 *              TCP/IP connection.
 *
 * INPUTS: none.
 *
 * OUTPUTS: char *
 *
 *****************************************************************************/

BOOLEAN connectAt(const char *machine, int32 port, int *sd)
{
  char thisHost[80];
#ifndef VXWORKS
  gethostname(thisHost, 80);
#else
  thisHost[0]='\0';
#endif
#ifndef NO_UNIX_SOCKETS
  /* If the names are the same, then check for the socket. */
  if ((strncmp(thisHost,machine,strlen(machine)) == 0) &&
      ((strlen(machine) == strlen(thisHost)) ||
       ((strlen(machine) < strlen(thisHost)) &&
	(thisHost[strlen(machine)] = '.'))))
    { /* Can try to connect at the unix socket. */
      if (!connectAtSocket(machine, port, sd)) {
	return(connectAtPort(machine, port, sd));
      } else {
	return TRUE;
      }
  } else {
    /* Have to connect at the TCP/IP socket. */
    return(connectAtPort(machine, port, sd));
  }
#else
    /* Have to connect at the TCP/IP socket. */
    return(connectAtPort(machine, port, sd));
#endif
}

/**********************************************************************
 *
 *  FUNCTION:  BOOLEAN listenAtPort(int port, int *sd)
 *
 *  DESCRIPTION: 
 *
 *		Create a socket for listening for new connections at a given
 * port value. If the value of port is 0, return the value chosen in port.
 *
 * Return 1 on success, else 0.
 *
 * NOTES:
 * 7-Jun-91: fedor: may need some changes for vxworks 
 *
 *********************************************************************/

BOOLEAN listenAtPort(int32 *port, int *sd)
{
  int reuse;
  int32 length;
  struct sockaddr_in server;
  
  bzero((char *)&server, sizeof(struct sockaddr_in));
  
  server.sin_port = htons(*port);
  
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  
  if ((*sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    tcaModWarning("socket fail\n");
    return FALSE;
  }
  
  reuse = 1;
  
#ifndef applec
  if ((setsockopt(*sd, SOL_SOCKET, SO_REUSEADDR,
		  (char *)&reuse, sizeof(int))) < 0) {
    tcaModWarning("setsockopt fail\n");
    return FALSE;
  }
#endif /* applec */
  
  reuse = 1;
  
#if !defined(VXWORKS) && !defined(applec)
  { int value = TCP_SOCKET_BUFFER;
    if ((setsockopt(*sd, IPPROTO_TCP, TCP_NODELAY,
		    (char *)&value, sizeof(int))) < 0)
      return FALSE;
    setsockopt(*sd, SOL_SOCKET, SO_SNDBUF, (char *)&value, sizeof(int));
    setsockopt(*sd, SOL_SOCKET, SO_RCVBUF, (char *)&value, sizeof(int));
  }
#endif /* !VXWORKS && !applec */
  
  if ((bind(*sd, (struct sockaddr *)&server, 
	    sizeof(struct sockaddr_in))) < 0)
    {
#if defined(VXWORKS)
      printErr("bind fail\n");
      printErrno(errno);
#endif
      return FALSE;
    }
  
  if (listen(*sd, BACKLOG) < 0) {
    tcaModWarning( "listen fail\n");
    return FALSE;
  }
  
  /* find what port is if not set */
  if (!*port) {
    length = sizeof(server);
    if (getsockname(*sd, (struct sockaddr *) &server, &length) < 0) {
      tcaModWarning( "getsockname fail\n");
      return FALSE;
    }
    
    *port = ntohs(server.sin_port);
  }
  
  return TRUE;
}

#ifndef NO_UNIX_SOCKETS
/**********************************************************************
 *
 *  FUNCTION:  BOOLEAN listenAtSocket(int port, int *sd)
 *
 *  DESCRIPTION: 
 *
 *		Create a socket for listening for new connections at a given
 * port value. If the value of port is 0, return the value chosen in port.
 *
 * Return 1 on success, else 0.
 *
 * NOTES:
 * 7-Jun-91: fedor: may need some changes for vxworks 
 *
 *********************************************************************/

BOOLEAN listenAtSocket(int port, int *sd)
{
  int reuse;
  int result;
  int oldUmask;
  struct sockaddr_un unix_server;
  
  bzero((char *)&unix_server, sizeof(struct sockaddr_un));
  unix_server.sun_family = AF_UNIX;
  sprintf(unix_server.sun_path,UNIX_SOCKET_NAME,port);
  
  /* Get rid of old links. */
  result = unlink(unix_server.sun_path);
  
  if (((result) != 0) &&
      ((errno) != ENOENT))
    tcaModWarning("Old socket %s could not be deleted. \n Is server still running?\n",
		  unix_server.sun_path);
  
  oldUmask = umask(!(S_IRUSR | S_IWUSR | S_IXUSR | 
		     S_IRGRP | S_IWGRP | S_IXGRP |
		     S_IROTH | S_IWOTH | S_IXOTH));
  
  /* Create directory. */
  result = mkdir(UNIX_DIR_NAME,
		 (S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP |
		  S_IROTH | S_IWOTH | S_IXOTH));
  if (((result) != 0) &&
      ((errno) != EEXIST))
    tcaModWarning("Can not create directory (%s)to hold socket connections.\n",
		  UNIX_DIR_NAME);
  
  if ((*sd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
    tcaModWarning("socket fail\n");
    return FALSE;
  }
  
  /* Change the mode of the socket so it can be deleted by others if needed.
   * This is safe since we alway try to do an tcp/ip create first and only
   * try and unlink the socket if the tcp/ip connection succeeds.
   */
  fchmod(*sd,(S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP |
	     S_IROTH | S_IWOTH | S_IXOTH));
  
  oldUmask = umask(oldUmask);

  reuse = 1;
  
#if !defined(VXWORKS) && !defined(applec)
  { int value = UNIX_SOCKET_BUFFER;
    setsockopt(*sd, SOL_SOCKET, SO_SNDBUF, (char *)&value, sizeof(int));
    setsockopt(*sd, SOL_SOCKET, SO_RCVBUF, (char *)&value, sizeof(int));
  }
#endif /* !VXWORKS && !applec */
  
  if ((bind(*sd, (struct sockaddr *)&unix_server,
	    sizeof(struct sockaddr_in))) < 0)
    {
#if defined(VXWORKS)
      printErr("bind fail\n");
      printErrno(errno);
#endif
      return FALSE;
    }
  
  if (listen(*sd, BACKLOG) < 0) {
    tcaModWarning( "listen fail\n");
    return FALSE;
  }
  
  return TRUE;
}

/**********************************************************************
 *
 *  FUNCTION:  BOOLEAN closeSocket(int port)
 *
 *  DESCRIPTION: 
 *
 *		Create a socket for listening for new connections at a given
 * port value. If the value of port is 0, return the value chosen in port.
 *
 * Return 1 on success, else 0.
 *
 * NOTES:
 * 7-Jun-91: fedor: may need some changes for vxworks 
 *
 *********************************************************************/

void closeSocket(int port)
{
  struct sockaddr_un unix_server;
  
  bzero((char *)&unix_server, sizeof(struct sockaddr_un));
  unix_server.sun_family = AF_UNIX;
  sprintf(unix_server.sun_path,UNIX_SOCKET_NAME,port);
  
  /* Get rid of old links. */
  unlink(unix_server.sun_path);
}
#endif /* NO_UNIX_SOCKETS */

/***************************************************************************/



/******************************************************************************
 *
 * FUNCTION int connectSocket(servHost, sd)
 *
 * DESCRIPTION: 
 * Returns TRUE if a connection was successfully opened to servHost, 
 * FALSE otherwise. sd is set to the value of the opened connection.
 *
 * INPUTS: 
 * char *servHost;
 * int *sd;
 *
 * OUTPUTS: int (TRUE or FALSE)
 *
 *****************************************************************************/

#if defined(__TURBOC__) || defined(__svr4__)
static char *index(char *str, char find)
{
	char *returnVal;
	for(returnVal =str; (*returnVal != find); returnVal++)
	 if(*returnVal == '\0') return NULL;
   return returnVal;
}
#endif

BOOLEAN connectSocket(const char *servHost, int *sd)
{ 
  char serverHostName[80], *colon;
  int serverPort;
  
  colon = index(servHost,':');
  if(colon != NULL) {
    /* servHost is <hostName>:<port> */
    serverPort = atoi(colon+1);
    /* Copy that portion of the name which is the real host name */
    strncpy(serverHostName, servHost, strlen(servHost)-strlen(colon));
    return connectAt(serverHostName, serverPort, sd);
  } else {
    return connectAt(servHost, SERVER_PORT, sd);
  }
}
