/*
 * File: parser.c
 * Author: Domingo Gallardo, CMU
 * Purpose: Parser of log messages. Translates messages to
 *          a structure named MsgData (see file msgData.h), 
 *          which contains the message's data.
 * 
 *
 * REVISION HISTORY
 *
 * $Log: Parser.c,v $
 * Revision 1.4  1995/05/31  20:58:35  rich
 * Fixed conflict with tca declarations.
 *
 * Revision 1.3  1995/04/19  14:32:00  rich
 * Added int32 for dealing with tca identifiers.
 *
 * Revision 1.2  1995/04/07  05:07:01  rich
 * Fixed GNUmakefiles to find the release directory.
 * Fixed problems found by sgi cc compiler.  It would not compile.
 *
 * Revision 1.1  1995/04/05  18:31:19  rich
 * Moved tview files to a subdirectory.
 *
 * Revision 1.2  1995/01/25  00:04:28  rich
 * Release of tca 7.9.  Mostly speed improvements.
 * The cvs binaries may now be located in /usr/local.
 * Formatting changes.
 *
 * Revision 1.1  1994/05/31  03:26:22  rich
 * Moved Domingo's tview tool into the main tca module.
 *
 * Revision 1.4  1994/05/27  05:34:48  rich
 * Can now read from file that is being written to.
 * Fixed Menu and button names.
 * Fixed Indentation.
 * Added menu item to change orientation of the tree.
 *
 * Revision 1.3  1993/09/07  00:24:43  domingo
 * Fixed almost all the warnings
 *
 * Revision 1.2  1993/08/13  02:09:24  domingo
 * Updated function declarations for compilation under gcc in ANSI C (still
 * a lot of warnings to fix).
 * Added automatic logging.
 *
 * Jan 1 1993 - Domingo Gallardo at School of Computer Science, CMU
 * Revised parser functions to take into account that the strings in msgData
 * are not initalized. Now ParseName & ParseModuleName return a just created
 * string.
 *
 * Dec 25 1992 - Domingo Gallardo at School of Computer Science, CMU
 * Created.
 *
 */

#include "Standard.h"
#include "Parser.h"
#include "Debug.h"
#include "Memory.h"
#include "MyString.h"
#include "TaskTree.h"

#define DEBUG_PARSER False

BOOLEAN _process_info_msg = FALSE;

/*
 *  Private functions
 */


static BOOLEAN IsInfoMsg(String name);
static msgType ParseType(StrMessage msg);
static String ParseName(StrMessage msg);
static int ParseId(StrMessage msg);
static String ParseModuleName(StrMessage msg);
static msgStatus ParseStatus(StrMessage msg);
static void ParseTime(StrMessage msg, MsgData data);
static int ParseIntervalOp(StrMessage msg);
static int ParseInterval(StrMessage msg);
static char ParsePointTC(StrMessage msg);
static int ParseTemporalConstraint(StrMessage msg);
static int ParseSituation(StrMessage msg);
static MsgData CreateMsgData(void);
void PrintMsgData(MsgData data);

/*
 *  Public functions
 */


MsgData ParseMsg(String str)
{
  MsgData data;
  _strMessage msg_aux, *msg;;
  
  data = CreateMsgData();
  
  
  msg_aux.message = msg_aux.str = str;
  msg = &msg_aux;
  data->type = ParseType(msg);
  switch (data->type) {
  case QUERY:
  case GOAL:
  case COMMAND:
  case INFORM:
  case EXCEPTION:
  case INTERVAL_MONITOR:
  case POLLING_MONITOR:
  case REPLY:
    data->name = ParseName(msg);
    data->id = ParseId(msg);
    msg->str = SkipUntilSubString(msg->str, ":");
    data->source = ParseModuleName(msg);
    data->parent_id = ParseId(msg);
    msg->str = SkipUntilSubString(msg->str, "-->");
    data->dest = ParseModuleName(msg);
    data->status = ParseStatus(msg);
    ParseTime(msg, data);
    break;
  case FIRE_DEMON:
    data->id = ParseId(msg);
    data->name = strdup("");
    msg->str = SkipUntilSubString(msg->str, ":");
    data->source = ParseModuleName(msg);
    data->parent_id = ParseId(msg);
    msg->str = SkipUntilSubString(msg->str, "-->");
    data->dest = ParseModuleName(msg);
    data->status = ParseStatus(msg);
    ParseTime(msg, data);
    break;
  case FAILURE:
    data->name = ParseName(msg);
    data->id = ParseId(msg);
    msg->str = SkipUntilSubString(msg->str, ":");
    /* Now it comes the name of the exception */
    data->source = ParseName(msg);
    data->parent_id = ParseId(msg);
    ParseTime(msg, data);
    break;
  case KILLED:
  case SUCCESS:
    data->name = ParseName(msg);
    data->id = ParseId(msg);
    msg->str = SkipUntilSubString(msg->str, ":");
    ParseTime(msg, data);
    break;
  case POINT_CONSTRAINT:
    data->point1 = ParseIntervalOp(msg);
    data->interval1 = ParseInterval(msg);
    data->msg1 = ParseName(msg);
    data->id1 = ParseId(msg);
    data->point_TC = ParsePointTC(msg);
    data->point2 = ParseIntervalOp(msg);
    data->interval2 = ParseInterval(msg);
    data->msg2 = ParseName(msg);
    data->id2 = ParseId(msg);
    data->name = strdup("");
    break;
  case TEMP_CONSTRAINT:
    msg->str = SkipSpaces(msg->str);
    data->temporal_constraint = ParseTemporalConstraint(msg);
    data->msg1 = ParseName(msg);
    data->id1 = ParseId(msg);
    msg->str = SkipUntilSubString(msg->str, "and");
    data->msg2 = ParseName(msg);
    data->id2 = ParseId(msg);
    data->name = strdup("");
    break;
  case WILL_KILL:
    msg->str = SkipUntilSubString(msg->str, "kill");
    data->name = ParseName(msg);
    data->id = ParseId(msg);
    data->kill_when = ParseSituation(msg);
    break;
  case RETRY:
    data->name = ParseName(msg);
    data->id = ParseId(msg);
    break;
  case WARNING:
    data->name = strdup("");
  default:
    Debug(DEBUG_PARSER, "Warning: unknown type >> %s\n", msg->message);
    break;
  }
  if (data->type != TYPE_NULL) {
    if (!_process_info_msg)
      if (IsInfoMsg(data->name))
	data->type = TYPE_NULL;
  }
  return (data);
}


/*
 *  PRIVATE FUNCTIONS
 */


static BOOLEAN IsInfoMsg(String name)
{
  static String infoMsgNames[] = {
    "msgInfoMsg",
    "classInfoMsg",
    "registerMessageMsg",
    "tapMessage",
    "registerHandlerMsg",
    "newModuleConnectMsg",
    "registerNamedFormMsg",
    "addHndToResourceMsg",
    "regMonitorMsg",
    "registerResourceMsg",
    "addExcepMsg",
    "createRefMsg",
    "referenceStatusMsg",
    NULL
  };
  
  int index = 0;
  
  while(infoMsgNames[index] != NULL)
    if (StringEqual(infoMsgNames[index++], name))
      return TRUE;
  return FALSE;
}



/*	Function Name: 
 *	Arguments:
 *	Description:
 *	Returns: 
 */


static msgType ParseType(StrMessage msg)
{
  
  static _msgTypeRec  type_table[] = {
    {"Query",        QUERY},
    {"Goal",         GOAL},
    {"Command",      COMMAND},
    {"Inform",       INFORM},
    {"Constraint",   INFORM},
    {"Exception",    EXCEPTION},
    {"Killed",       KILLED},
    {"IMonitor",     INTERVAL_MONITOR},
    {"PMonitor",     POLLING_MONITOR},
    {"FireDemon",    FIRE_DEMON},
    {"  Retry",      RETRY},
    {"  Reply",      REPLY},
    {"  Failure",    FAILURE},
    {"  Success",    SUCCESS},
    {"  Done",       SUCCESS},
    {"  ByPass",     BYPASS},
    {"   Asserting", POINT_CONSTRAINT},
    {"Will",         WILL_KILL},
    {"Adding",       TEMP_CONSTRAINT},
    {"WARNING:",     WARNING},
  };
  
  int index;
  char *my_string;
  
  my_string = msg->str;
  for (index = 0; index <  sizeArray(type_table, _msgTypeRec); index++)
    if (BeginsWith(my_string, type_table[index].string)) {
      msg->str = msg->str + strlen(type_table[index].string);
      return(type_table[index].type);
    }
  return (TYPE_NULL);
}

static String ParseName(StrMessage msg)
{
  String name, aux;
  char aux_char;
  
  msg->str = SkipSpaces(msg->str);
  aux = msg->str;
  while(isalpha(*msg->str) ||
	*msg->str == '_' ||
	*msg->str == '-')
    msg->str++;
  aux_char = *msg->str;
  *msg->str = '\0';
  name = strdup(aux);
  *msg->str = aux_char;
  return(name);
}

static int ParseId(StrMessage msg)
{
  int32 msg_id;
  String aux_str;
  
  aux_str = SkipUntilSubString(msg->str, "{");
  if (aux_str != msg->str){
    sscanf(aux_str, "%d", &msg_id);
    msg->str = SkipUntilSubString(aux_str, "}");
    return (msg_id == -1 ? ROOT_NODE_ID : msg_id);
  }
  return(-1);
}


/*	Function Name: ParseModuleName
 *	Arguments: - msg: the String being processing
 *                 - module_name: the string where the copy is done
 *	Description: Copies in module_name the substring with the 
 *                   chars: "*", " " or alphabetic. Once ended trims
 *                   all the spaces in the right side of module_name. 
 *                   Leaves the pointer in the first character that 
 *                   not satisfy the former condition.
 *	Returns: Nothing
 */

static String ParseModuleName(StrMessage msg)
{
  String initial_str, aux_str, return_str;
  
  aux_str = msg->str;
  aux_str = SkipSpaces(aux_str);
  initial_str = aux_str;
  if (!isalpha(*aux_str) && 
      *aux_str != '*') {
    Debug(DEBUG_PARSER, "Warning: No module name found in >> %s\n", msg->str);
  }
  else {
    while ((isalpha(*aux_str)) ||
	   (*aux_str == '*') ||
	   (*aux_str == ' ')) {
      aux_str++;
    }
  }
  aux_str--;
  while (*aux_str == ' ')
    --aux_str;
  aux_str++;
  *aux_str = '\0';
  return_str = strdup(initial_str);
  *aux_str = ' ';
  msg->str = aux_str;
  return (return_str);
}

static msgStatus ParseStatus(StrMessage msg)
{
  
  static _msgStatusRec status_table[] = {
    {"Sent",          SENT},
    {"Pending",       PENDING},
    {"Inactive",      INACTIVE},
    {"Completed",     COMPLETED},
  };
  
  int index;
  String aux_msg;
  
  msg->str = SkipSpaces(msg->str);
  aux_msg = msg->str;
  aux_msg = SkipUntilSubString(aux_msg, "(");
  if (aux_msg != msg->str) {
    for (index = 0; index < sizeArray(status_table, _msgStatusRec); index++)
      if (BeginsWith(aux_msg, status_table[index].string)) {
	msg->str = msg->str + strlen(status_table[index].string) + 1;
	return(status_table[index].type);
      }
  }
  Debug(DEBUG_PARSER, "Warning: unknown status in >> %s\n", msg->str);
  return (STATUS_NULL);
}

static void ParseTime(StrMessage msg, MsgData data)
{
  msg->str = SkipSpaces(msg->str);
  sscanf(msg->str, "%d", &(data->hours));
  msg->str = SkipAlphaNum(msg->str);
  msg->str = SkipUntilSubString(msg->str, ":");
  sscanf(msg->str, "%d", &(data->minutes));
  msg->str = SkipUntilSubString(msg->str, ":");
  sscanf(msg->str, "%d", &(data->seconds));
  msg->str = SkipUntilSubString(msg->str, ".");
  sscanf(msg->str, "%d", &(data->m_seconds));
  data->time = 360000 * data->hours + 6000 * data->minutes +
    100 * data->seconds + data->m_seconds;
}


static int ParseIntervalOp(StrMessage msg)
{
  msg->str = SkipSpaces(msg->str);
  switch (*msg->str) {
  case 'S':
    msg->str = SkipUntilSubString(msg->str, "Start(");
    return(START);
  case 'E':
    msg->str = SkipUntilSubString(msg->str, "End(");
    return(END);
  default:
    return(0);
  }
}


static int ParseInterval(StrMessage msg)
{
  msg->str = SkipSpaces(msg->str);
  switch(*msg->str) {
  case 'P':
    msg->str = SkipUntilSubString(msg->str, "Planning(");
    return(PLANNING_INTERVAL);
  case 'A':
    msg->str = SkipUntilSubString(msg->str, "Achievement(");
    return(ACHIVEMENT_INTERVAL);
  case 'H':
    msg->str = SkipUntilSubString(msg->str, "Handling(");
    return(HANDLING_INTERVAL);
  default:
    return(0);
  }
}


static char ParsePointTC(StrMessage msg)
{
  msg->str = SkipSpaces(msg->str);
  switch (*msg->str) {
  case '=':
    msg->str++;
    return ('=');
  case '<':
    msg->str++;
    if (*msg->str == '=') {
      msg->str++;
      return ('\\');           /* <= */
    }
    else
      return ('<');
  case '>':
    msg->str++;
    if (*msg->str == '=') {
      msg->str++;
      return ('/');           /* >= */
    }
    else
      return ('>');
  }
  return(' ');
}


static int ParseTemporalConstraint(StrMessage msg)
{
  
  static _msgTCRec TC_table[] = {
    {"SEQUENTIAL ACHIEVEMENT constraint between", MSG_SEQ_ACH},
    {"SEQUENTIAL PLANNING constraint between", MSG_SEQ_PLAN},
    {"DELAY PLANNING constraint to", MSG_DELAY_PLAN},
    {"PLAN FIRST constraint to", MSG_PLAN_FIRST},
  };
  
  int index;
  char *my_string;
  
  my_string = msg->str;
  for (index = 0; index <  sizeArray(TC_table, _msgTCRec); index++)
    if (BeginsWith(my_string, TC_table[index].string)) {
      msg->str = msg->str + strlen(TC_table[index].string);
      return(TC_table[index].type);
    }
  return (TYPE_NULL);
}

static int ParseSituation(StrMessage msg)
{
  
  static _msgKillSituationRec killSituation_table[] = {
    {"when it completes", WHEN_COMPLETES},
    {"when submessages complete", WHEN_SUBMESSAGES_COMPLETES},
    {"when all references to it are released", WHEN_REFERENCES_RELEASED}
  };
  
  int index;
  char *my_string;
  
  msg->str = SkipSpaces(msg->str);
  my_string = msg->str;
  for (index = 0; index <  sizeArray(killSituation_table,
				     _msgKillSituationRec); index++)
    if (BeginsWith(my_string, killSituation_table[index].string)) {
      msg->str = msg->str + strlen(killSituation_table[index].string);
      return(killSituation_table[index].type);
    }
  Warning(">> Warning in ParseSituation: %s is not a known kill situation\n",
	  my_string);
  return (NULL_KILL_SITUATION);
}


static MsgData CreateMsgData(void)
{
  MsgData data;
  
  data = _new(_MsgData);
  data->type = TYPE_NULL;
  data->id = -1;
  data->status = STATUS_NULL;
  data->time = 0;
  data->hours = 0;
  data->minutes = 0;
  data->seconds = 0;
  data->m_seconds = 0;
  data->msg1 = NULL;
  data->msg2 = NULL;
  return(data);
}


void PrintMsgData(MsgData data)
{
  printf("Type: %d\nName: %s\nId: %d\nSource: %s\nDest: %s\nStatus: %d\nTime: %d (%d:%d:%d.%d)\n",
	 data->type, data->name, data->id,
	 data->source, data->dest, data->status,
	 data->time, data->hours, data->minutes, data->seconds,
	 data->m_seconds);
}


/* ------------------------------------------------ */
/*                   MODULE TEST                    */
/* ------------------------------------------------ */


/* ------------------------------------------------
   
   main()
   {
   static  char *msg;
   MsgData data;
   
   msg = strdup("Query        classInfoMsg {0}: **NEWMOD** --> TCA Server      (Sent)  0:28:52.68");
   data = ParseMsg(msg);
   PrintMsgData(data);
   printf("\n");
   msg = strdup("Query        classInfoMsg : **NEWMOD** --> TCA Server      (Sent)  0:28:52.68");
   data = ParseMsg(msg);
   PrintMsgData(data);
   printf("\n");
   msg = strdup("Query        classInfoMsg : **NEWMOD** -->       (Sent)  0:28:52.68");
   data = ParseMsg(msg);
   PrintMsgData(data);
   printf("\n");
   }
   
   
   ------------------------------------------------ */
