/*****************************************************************************
 * PROJECT: TCA
 *
 * (c) Copyright Richard Goodwin, 1995. All rights reserved.
 *
 * FILE: tcaDev.c
 *
 * ABSTRACT:
 * 
 * This file provides routines for connecting TCA to the set of device
 * interfaces (see devUtils.{c,h}).
 *
 * ADAPTED FROM XAVIER SOFTWARE TO FACILITATE ADDING I/O TO MODULES.
 *
 * $Source: /afs/cs.cmu.edu/project/TCA/Master/tcaV8/utils/tcaDev.c,v $ 
 * $Revision: 1.8.2.1 $
 * $Date: 1995/10/10 00:33:02 $
 * $Author: rich $
 *
 * REVISION HISTORY:
 *
 * $Log: tcaDev.c,v $
 * Revision 1.8.2.1  1995/10/10  00:33:02  rich
 * Fixed a bug that caused a module using tcaDev.h to crash if central was
 * not running first.
 *
 * Revision 1.8  1995/08/14  21:35:04  rich
 * Changed devReadN and devWriteN to take the same parameters as read and
 * write.  This allows macro substitution for the system calls.
 * Added "dev" prefix to everything.  Some minor fixes.
 *
 * Revision 1.7  1995/08/05  23:23:36  rich
 * Added functional interface to devUtils.  See the README for details.
 *
 * Revision 1.6  1995/07/19  14:32:57  rich
 * Error checking on select and accept calls.
 *
 * Revision 1.5  1995/06/14  03:26:15  rich
 * Added DBMALLOC_DIR.
 * Fixed problems with multiple connections.
 *
 * Revision 1.4  1995/06/01  00:28:28  rich
 * Use TIME_MAX rather than LONG_MAX for timeouts.
 * Fix some problems with devUtils processOutput.
 *
 * Revision 1.3  1995/04/07  05:11:21  rich
 * Fixed GNUmakefiles to find the release directory.
 * Moved all system includes into libc.h
 * Moved global variables into the c files and got rid of #define DECLARE_...
 * Now works with xavier stuff.
 *
 * Revision 1.2  1995/04/04  19:48:25  rich
 * Added sgi support.
 * Changed setAlarm to setTimer.
 * Numerous improvements and bug fixes in devUtils.
 *
 * Revision 1.1  1995/03/30  15:53:34  rich
 * DBMALLOC works.  To use "gmake -k -w DBMALLOC=DBMALLOC install"
 * Added ezx library and the devutils library.
 *
 *
 * 30-Jan-1993 Richard Goodwin Created.
 *
 *****************************************************************************/

#include "tca/libc.h"
#include "tca.h"

#include "devUtils.h"
#include "tcaDev.h"

/*****************************************************************************
 * Global constants
 *****************************************************************************/

/*****************************************************************************
 * Global variables
 *****************************************************************************/

TCA_PTR    TCA_device = NULL;
/*
{
  { 
    NOT_CONNECTED,
    FALSE,
    { "", DEFAULT_PORT},
    { "TCA central", DEFAULT_BAUD},
    "TCA central",
    NO_FDS,
    NO_FDS,
    LISTENING | TALKING | ACCEPTING,
    FALSE,
    (FILE *) NULL,
    &devConnections,
    (DEVICE_OUTPUT_HND) TCA_outputHnd,
    Null_Handler,  
    {TIME_MAX, 0},
    Null_Handler,
    {0, 0},
    {TIME_MAX, 0},
    (void (*)(DEV_PTR)) NULL,  
    (void (*)(DEV_PTR)) NULL,  
    (void (*)(DEV_PTR)) NULL,  
  },
  (char *) NULL,
  (void (*)(void)) NULL,
  (void (*)(void)) NULL,
  (void (*)(void)) NULL,
  (const char **)NULL,
  (const char **)NULL
};
*/

/*****************************************************************************
 * Forward Declarations
 *****************************************************************************/

void tcaExitHnd(void);

/*****************************************************************************
 *
 * FUNCTION: void TCA_outputHnd(int fd, long chars_available)
 *
 * DESCRIPTION: Handles TCA events.
 *
 * INPUTS:
 *
 * OUTPUTS:
 *
 * HISTORY:
 *
 *****************************************************************************/

void TCA_outputHnd(int fd, long chars_available)
{
  if (tcaHandleMessage(0) != Success) {
    if (fd_isZero(tcaGetConnections()))
      devDisconnectDev(TCA_device->dev,FALSE,TRUE);
  }
  devUpdateConnections(TCA_device->dev,tcaGetConnections());
}

static void try_connect(void)
{
  tcaRegisterExitProc(tcaExitHnd);
  if (tcaServerMachine() != NULL) {
    tcaConnectModule(TCA_device->moduleName, tcaServerMachine());
  } else {
    static char thisHost[DEFAULT_LINE_LENGTH+1];

    gethostname(thisHost,DEFAULT_LINE_LENGTH);
    fprintf(stderr,"CENTRALHOST environment variable not set, using %s\n",
	    thisHost);
    tcaConnectModule(TCA_device->moduleName, thisHost);
  }
  
  if (tcaGetServer() > 0) {
    devConnectDev(TCA_device->dev,tcaGetServer());
    (* TCA_device->registerAll)();
    tcaEnableDistributedResponses();
    tcaModuleProvidesArgv(TCA_device->provides);
    tcaModuleRequiresArgv(TCA_device->requires);
    tcaWaitUntilReady();
    devUpdateConnections(TCA_device->dev,tcaGetConnections());
  }
}

void TCA_disconnect(DEV_PTR ignore)
{
  if (TCA_device->disconnectHnd != NULL)
    (* TCA_device->disconnectHnd)();
  tcaClose();
  tcaRegisterExitProc(tcaExitHnd);
  devDisconnectDev(TCA_device->dev,TRUE,TRUE);
}

void TCA_reconnect(DEV_PTR ignore)
{
  try_connect();
  if (TCA_isConnected())
    if (TCA_device->reconnectHnd != NULL)
      (* TCA_device->reconnectHnd)();
}


void TCA_connect(char *moduleName,void registerAll(void),
		 void disconnectFn(void),void reconnectFn(void),
		 const char **provides, const char **requires)
{
  TCA_device = (TCA_PTR)malloc(sizeof(TCA_TYPE));
  TCA_device->moduleName = moduleName;
  TCA_device->registerAll = registerAll;
  TCA_device->disconnectHnd = disconnectFn;
  TCA_device->reconnectHnd = reconnectFn;
  TCA_device->provides = provides;
  TCA_device->requires = requires;
  
  TCA_device->dev = 
    devCreateDev("TCA device",
		 DEV_OUTPUTHND, TCA_outputHnd,
		 DEV_LISTENING, LISTENING | TALKING | ACCEPTING,
		 DEV_DISCONNECTHND, TCA_disconnect,
		 DEV_RECONNECTHND, TCA_reconnect,
		 NULL);
  tcaRegisterExitProc(tcaExitHnd);
  TCA_device->moduleName = moduleName;
  TCA_device->registerAll = registerAll;
  TCA_device->disconnectHnd = disconnectFn;
  TCA_device->reconnectHnd = reconnectFn;

  TCA_device->provides = provides;
  TCA_device->requires = requires;

  try_connect();
}

void tcaExitHnd(void)
{
  /* prevent the disconnectin handler from being reentered. */
  static BOOLEAN inTCAExitHnd=FALSE;

  if (inTCAExitHnd)
    return;
  inTCAExitHnd = TRUE;
  fprintf(stderr,"TCA would exit here\n");
  if (tcaGetServer() <= 0) {
    devDisconnectDev(TCA_device->dev,TRUE,TRUE);
  }
  inTCAExitHnd = FALSE;
}

BOOLEAN TCA_isConnected(void)
{
  return devIsConnected(TCA_device->dev);
}


BOOLEAN TCA_isDisconnected(void)
{
  return devIsDisconnected(TCA_device->dev);
}
